﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using HeuristicLab.Analysis;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Common.Resources;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.MainForm;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Persistence.Default.Xml;
using RunCreationClient = HeuristicLab.Clients.OKB.RunCreation.RunCreationClient;
using SingleObjectiveOKBProblem = HeuristicLab.Clients.OKB.RunCreation.SingleObjectiveOKBProblem;

namespace HeuristicLab.OptimizationExpertSystem {
  [Item("Expert-System", "Currently in experimental phase, an expert system that makes algorithm suggestions based on fitness landscape analysis features and an optimization knowledge base.")]
  [StorableClass]
  [Creatable(CreatableAttribute.Categories.TestingAndAnalysis, Priority = 119)]
  public sealed class ExpertSystem : NamedItem, IStorableContent, INotifyPropertyChanged {

    public string Filename { get; set; }

    public static new Image StaticItemImage {
      get { return VSImageLibrary.Library; }
    }

    [Storable]
    private int maximumEvaluations;
    public int MaximumEvaluations {
      get { return maximumEvaluations; }
      set {
        if (maximumEvaluations == value) return;
        maximumEvaluations = value;
        OnPropertyChanged("MaximumEvaluations");
        UpdateSuggestions();
      }
    }

    [Storable]
    private RunCollection runs;
    public RunCollection Runs {
      get { return runs; }
    }

    [Storable]
    private RunCollection knowledgeBase;
    public RunCollection KnowledgeBase {
      get { return knowledgeBase; }
      set {
        if (knowledgeBase == value) return;
        knowledgeBase = value;
        OnPropertyChanged("KnowledgeBase");
      }
    }

    [Storable]
    private SingleObjectiveOKBProblem problem;
    public SingleObjectiveOKBProblem Problem {
      get { return problem; }
      set {
        if (problem == value) return;
        problem = value;
        OnPropertyChanged("Problem");
        UpdateSuggestions();
      }
    }

    [Storable]
    private ItemList<IAlgorithm> suggestedInstances;
    private ReadOnlyItemList<IAlgorithm> readOnlySuggestedInstances;
    public ReadOnlyItemList<IAlgorithm> SuggestedInstances {
      get { return readOnlySuggestedInstances; }
    }

    [Storable]
    private RunCollection problemInstances;
    public RunCollection ProblemInstances {
      get { return problemInstances; }
      set {
        if (problemInstances == value) return;
        problemInstances = value;
        OnPropertyChanged("ProblemInstances");
      }
    }

    [Storable]
    private BidirectionalLookup<long, IRun> algorithmId2RunMapping;
    [Storable]
    private BidirectionalDictionary<long, IAlgorithm> algorithmId2AlgorithmInstanceMapping;

    private bool Maximization {
      get {
        return Problem != null && Problem.ProblemId >= 0 && ((IValueParameter<BoolValue>)Problem.MaximizationParameter).Value.Value;
      }
    }

    [StorableConstructor]
    private ExpertSystem(bool deserializing) : base(deserializing) { }
    private ExpertSystem(ExpertSystem original, Cloner cloner)
      : base(original, cloner) {
      runs = cloner.Clone(original.runs);
      knowledgeBase = cloner.Clone(original.knowledgeBase);
      suggestedInstances = cloner.Clone(original.suggestedInstances);
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      problemInstances = cloner.Clone(original.problemInstances);
      problem = cloner.Clone(original.problem);
      algorithmId2RunMapping = new BidirectionalLookup<long, IRun>();
      foreach (var kvp in original.algorithmId2RunMapping.FirstEnumerable) {
        algorithmId2RunMapping.AddRangeFirst(kvp.Key, kvp.Select(cloner.Clone));
      }
      algorithmId2AlgorithmInstanceMapping = new BidirectionalDictionary<long, IAlgorithm>();
      foreach (var kvp in original.algorithmId2AlgorithmInstanceMapping) {
        algorithmId2AlgorithmInstanceMapping.Add(kvp.Key, cloner.Clone(kvp.Value));
      }
      RegisterEventHandlers();
    }
    public ExpertSystem() {
      Name = ItemName;
      Description = ItemDescription;
      runs = new RunCollection();
      knowledgeBase = new RunCollection();
      suggestedInstances = new ItemList<IAlgorithm>();
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      problemInstances = new RunCollection();
      problem = new SingleObjectiveOKBProblem();
      algorithmId2RunMapping = new BidirectionalLookup<long, IRun>();
      algorithmId2AlgorithmInstanceMapping = new BidirectionalDictionary<long, IAlgorithm>();
      RegisterEventHandlers();
    }

    private void ProblemOnProblemChanged(object sender, EventArgs eventArgs) {
      if (Problem == null) return;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ExpertSystem(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      RegisterEventHandlers();
    }

    private void RegisterEventHandlers() {
      problem.ProblemChanged += ProblemOnProblemChanged;
      runs.CollectionReset += InformationChanged;
      runs.ItemsAdded += InformationChanged;
      runs.ItemsRemoved += InformationChanged;
      runs.Reset += InformationChanged;
      runs.UpdateOfRunsInProgressChanged += InformationChanged;
      knowledgeBase.CollectionReset += InformationChanged;
      knowledgeBase.ItemsAdded += InformationChanged;
      knowledgeBase.ItemsRemoved += InformationChanged;
    }

    private void InformationChanged(object sender, EventArgs e) {
      var runCollection = sender as RunCollection;
      if (runCollection != null && runCollection.UpdateOfRunsInProgress) return;
      UpdateSuggestions();
    }

    public void UpdateInstanceMap() {
      var flaValues = ProblemInstances.ResultNames
                      .Where(x => x.StartsWith("Characteristic.")
                        && ProblemInstances.All(y => y.Results.ContainsKey(x) && y.Results[x] is DoubleValue))
                      .Select(x => x).ToList();
      var ds = new double[ProblemInstances.Count, flaValues.Count];
      var instanceCounter = 0;
      foreach (var instance in ProblemInstances) {
        for (var feature = 0; feature < flaValues.Count; feature++)
          ds[instanceCounter, feature] = ((DoubleValue)instance.Results[flaValues[feature]]).Value;
        instanceCounter++;
      }

      int info;
      double[] s2;
      double[,] v;
      alglib.pcabuildbasis(ds, instanceCounter, flaValues.Count, out info, out s2, out v);

      ProblemInstances.UpdateOfRunsInProgress = true;
      try {
        instanceCounter = 0;
        foreach (var instance in ProblemInstances) {
          double x = 0, y = 0;
          for (var feature = 0; feature < flaValues.Count; feature++) {
            x += ds[instanceCounter, feature] * v[feature, 0];
            y += ds[instanceCounter, feature] * v[feature, 1];
          }
          IItem item;
          if (instance.Results.TryGetValue("Projection.PCA1", out item)) {
            ((DoubleValue)item).Value = x;
          } else instance.Results.Add("Projection.PCA1", new DoubleValue(x));
          if (instance.Results.TryGetValue("Projection.PCA2", out item)) {
            ((DoubleValue)item).Value = y;
          } else instance.Results.Add("Projection.PCA2", new DoubleValue(y));

          instanceCounter++;
        }
      } finally { ProblemInstances.UpdateOfRunsInProgress = false; }
    }

    private static readonly HashSet<string> InterestingValueNames = new HashSet<string>() {
      "QualityPerEvaluations", "Problem Name", "Problem Type", "Algorithm Name", "Algorithm Type", "Maximization", "BestKnownQuality"
    };

    public async void UpdateKnowledgeBaseAsync(IProgress progress) {
      progress.Start("Updating Knowledge Base from OKB");
      await Task.Factory.StartNew(() => { DoUpdateKnowledgeBase(progress); }, TaskCreationOptions.LongRunning);
    }

    public void UpdateKnowledgeBase() {
      DoUpdateKnowledgeBase(new Progress(string.Empty, ProgressState.Started));
    }

    private void DoUpdateKnowledgeBase(IProgress progress) {
      var queryClient = Clients.OKB.Query.QueryClient.Instance;
      var adminClient = Clients.OKB.Administration.AdministrationClient.Instance;
      try {
        progress.Status = "Downloading run information...";
        progress.ProgressValue = 0;
        // FIXME: How to tell if refresh is necessary?
        queryClient.Refresh();
        progress.ProgressValue = 0.5;
        progress.Status = "Downloading algorithm and problem instance information...";
        // FIXME: How to tell if refresh is necessary?
        adminClient.Refresh();

        var probInstance = adminClient.Problems.SingleOrDefault(x => x.Id == Problem.ProblemId);
        if (probInstance == null) throw new InvalidOperationException("The chosen problem instance cannot be found in the OKB.");
        var probClassId = probInstance.ProblemClassId;

        var problemClassFilter = (Clients.OKB.Query.StringComparisonAvailableValuesFilter)queryClient.Filters.Single(x => x.Label == "Problem Class Name");
        problemClassFilter.Value = adminClient.ProblemClasses.Single(x => x.Id == probClassId).Name;

        progress.Status = "Downloading problem instances...";
        progress.ProgressValue = 0;
        var p = 0;
        ProblemInstances.UpdateOfRunsInProgress = true;
        ProblemInstances.Clear();
        var totalProblems = adminClient.Problems.Count(x => x.ProblemClassId == probClassId);
        foreach (var problInst in adminClient.Problems.Where(x => x.ProblemClassId == probClassId)) {
          progress.Status = string.Format("Downloading problem {0} (okb-id: {1})....", problInst.Name, problInst.Id);
          var data = Clients.OKB.Administration.AdministrationClient.GetProblemData(problInst.Id);
          if (data != null) {
            using (var stream = new MemoryStream(data)) {
              try {
                var prob = (IProblem)XmlParser.Deserialize<IContent>(stream);
                var probRun = new Run() { Name = prob.Name };
                prob.CollectParameterValues(probRun.Parameters);
                progress.Status += Environment.NewLine + "Downloading characteristics...";
                foreach (var v in RunCreationClient.GetCharacteristicValues(problInst.Id)) {
                  probRun.Results.Add("Characteristic." + v.Name, RunCreationClient.Instance.ConvertToItem(v));
                }
                ProblemInstances.Add(probRun);
              } catch { }
              stream.Close();
            }
          }
          p++;
          progress.ProgressValue = p / (double)totalProblems;
        }

        algorithmId2AlgorithmInstanceMapping.Clear();
        progress.Status = "Downloading algorithm instances...";
        progress.ProgressValue = 0;
        p = 0;
        foreach (var algInst in adminClient.Algorithms) {
          progress.Status = string.Format("Downloading algorithm {0} (okb-id: {1})...", algInst.Name, algInst.Id);
          var data = Clients.OKB.Administration.AdministrationClient.GetAlgorithmData(algInst.Id);
          if (data != null) {
            using (var stream = new MemoryStream(data)) {
              try {
                var alg = (IAlgorithm)XmlParser.Deserialize<IContent>(stream);
                algorithmId2AlgorithmInstanceMapping.Add(algInst.Id, alg);
              } catch { }
              stream.Close();
            }
          }
          p++;
          progress.ProgressValue = p / (double)adminClient.Algorithms.Count;
        }

        var interestingValues = queryClient.ValueNames.Where(x => InterestingValueNames.Contains(x.Name)).ToList();

        progress.Status = "Obtaining number of runs...";
        progress.ProgressValue = 0;
        p = 0;
        var count = queryClient.GetNumberOfRuns(problemClassFilter);
        if (count == 0) return;

        var runIds = queryClient.GetRunIds(problemClassFilter).ToList();
        var conversions = new List<Task>();
        var runList = new List<IRun>();
        while (p < count) {
          var nextIds = runIds.Skip(p).Take(500).ToList();
          progress.Status = string.Format("Downloading runs {0} to {1} of {2}...", p, p + nextIds.Count, count);
          var okbRuns = queryClient.GetRunsWithValues(nextIds, true, interestingValues);
          conversions.Add(Task.Factory.StartNew(() => {
            var hlRuns = okbRuns.AsParallel().Select(x => new { AlgorithmId = x.Algorithm.Id, Run = queryClient.ConvertToOptimizationRun(x) }).ToList();
            lock (runList) {
              foreach (var r in hlRuns) {
                algorithmId2RunMapping.Add(r.AlgorithmId, r.Run);
                runList.Add(r.Run);
              }
            }
          }));
          p += nextIds.Count;
          progress.ProgressValue = p / (double)count;
        }
        Task.WaitAll(conversions.ToArray());

        progress.Status = "Finishing...";
        var algInstRunDict = runList.Where(x => x.Parameters.ContainsKey("Problem Name") && x.Parameters["Problem Name"] is StringValue)
                                          .GroupBy(x => ((StringValue)x.Parameters["Problem Name"]).Value)
                                          .ToDictionary(x => x.Key, x => x.GroupBy(y => ((StringValue)y.Parameters["Algorithm Name"]).Value)
                                                                                  .ToDictionary(y => y.Key, y => y.ToList()));

        foreach (var instance in ProblemInstances) {
          IItem probNameParam;
          if (!instance.Parameters.TryGetValue("Problem Name", out probNameParam)) continue;

          var probInstanceName = ((StringValue)probNameParam).Value;
          var bkQuality = ((DoubleValue)instance.Parameters["BestKnownQuality"]).Value;
          var maximization = ((BoolValue)instance.Parameters["Maximization"]).Value;

          if (!algInstRunDict.ContainsKey(probInstanceName)) continue;
          foreach (var kvp in algInstRunDict[probInstanceName]) {
            var algInstanceName = kvp.Key;
            // TODO: Things needs to be configurable here (table name, targets)
            foreach (var target in new[] { 1, 1.01, 1.05, 1.1, 1.2, 1.5 }) {
              var result = ExpectedRuntimeHelper.CalculateErt(kvp.Value, "QualityPerEvaluations", bkQuality * target, maximization);
              var resultName = algInstanceName + "@" + ((target - 1) * 100) + "%";
              IItem item;
              if (instance.Results.TryGetValue(resultName, out item)) {
                ((DoubleValue)item).Value = Math.Log10(result.ExpectedRuntime);
              } else instance.Results.Add(resultName, new DoubleValue(Math.Log10(result.ExpectedRuntime)));
            }
          }
        }
        KnowledgeBase = new RunCollection(runList);
      } finally { progress.Finish(); ProblemInstances.UpdateOfRunsInProgress = false; }
    }

    private void UpdateSuggestions() {
      if (Problem == null) return;
      var instances = new SortedList<double, IAlgorithm>();
      foreach (var relevantRuns in knowledgeBase.GroupBy(x => x.Algorithm)) {
        var avgQuality = 0.0;
        var counter = 0;
        foreach (var run in relevantRuns) {
          var performanceGraph = ((IndexedDataTable<double>)run.Results["QualityPerEvaluations"]);
          try {
            avgQuality += performanceGraph.Rows.First().Values.TakeWhile(x => x.Item1 < MaximumEvaluations).Last().Item2;
            counter++;
          } catch { continue; }
        }
        avgQuality /= counter;
        instances.Add(avgQuality, relevantRuns.Key);
      }

      suggestedInstances.Clear();
      var instanceLadder = instances.Select(x => (IAlgorithm)x.Value.Clone());
      suggestedInstances.AddRange(Maximization ? instanceLadder.Reverse() : instanceLadder);
    }

    public event PropertyChangedEventHandler PropertyChanged;
    private void OnPropertyChanged(string propertyName) {
      var handler = PropertyChanged;
      if (handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
    }
  }
}
