﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.QuadraticAssignment;

namespace HeuristicLab.Problems.CharacteristicAnalysis.QAP {
  [Item("QAP Characteristic Calculator", "")]
  [StorableClass]
  public sealed class QAPCharacteristicCalculator : CharacteristicCalculator {

    public override IEnumerable<string> Characteristics {
      get {
        return new[] { "Dimension",
        "FlowDominance", "DistanceDominance",
        "FlowAsymmetry", "DistanceAsymmetry",
        "FlowSparsity", "DistanceSparsity",
        "FlowSkewness", "DistanceSkewness",
        "FlowDisparity", "DistanceDisparity" };
      }
    }

    [StorableConstructor]
    private QAPCharacteristicCalculator(bool deserializing) : base(deserializing) { }
    private QAPCharacteristicCalculator(QAPCharacteristicCalculator original, Cloner cloner) : base(original, cloner) { }
    public QAPCharacteristicCalculator() { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new QAPCharacteristicCalculator(this, cloner);
    }

    public override bool CanCalculate(IProblem problem) {
      return problem is QuadraticAssignmentProblem;
    }

    public override IEnumerable<KeyValuePair<string, IItem>> Calculate(IProblem problem, IEnumerable<string> characteristics = null) {
      var qap = problem as QuadraticAssignmentProblem;
      if (qap == null) throw new ArgumentException("Instance is not of type QuadraticAssignmentProblem", "problem");
      var allowed = new HashSet<string>(characteristics ?? Characteristics);
      if (allowed.Contains("Dimension"))
        yield return new KeyValuePair<string, IItem>("Dimension", new Data.IntValue(qap.Weights.Rows));
      if (allowed.Contains("FlowDominance"))
        yield return new KeyValuePair<string, IItem>("FlowDominance", new Data.DoubleValue(DoubleMatrixCharacteristicCalculator.CoeffVariation(qap.Weights)));
      if (allowed.Contains("DistanceDominance"))
        yield return new KeyValuePair<string, IItem>("DistanceDominance", new Data.DoubleValue(DoubleMatrixCharacteristicCalculator.CoeffVariation(qap.Distances)));
      if (allowed.Contains("FlowAsymmetry"))
        yield return new KeyValuePair<string, IItem>("FlowAsymmetry", new Data.DoubleValue(DoubleMatrixCharacteristicCalculator.Asymmetry(qap.Weights)));
      if (allowed.Contains("DistanceAsymmetry"))
        yield return new KeyValuePair<string, IItem>("DistanceAsymmetry", new Data.DoubleValue(DoubleMatrixCharacteristicCalculator.Asymmetry(qap.Distances)));
      if (allowed.Contains("FlowSparsity"))
        yield return new KeyValuePair<string, IItem>("FlowSparsity", new Data.DoubleValue(DoubleMatrixCharacteristicCalculator.Sparsity(qap.Weights)));
      if (allowed.Contains("DistanceSparsity"))
        yield return new KeyValuePair<string, IItem>("DistanceSparsity", new Data.DoubleValue(DoubleMatrixCharacteristicCalculator.Sparsity(qap.Distances)));
      if (allowed.Contains("FlowSkewness"))
        yield return new KeyValuePair<string, IItem>("FlowSkewness", new Data.DoubleValue(DoubleMatrixCharacteristicCalculator.Skewness(qap.Weights)));
      if (allowed.Contains("DistanceSkewness"))
        yield return new KeyValuePair<string, IItem>("DistanceSkewness", new Data.DoubleValue(DoubleMatrixCharacteristicCalculator.Skewness(qap.Distances)));
      if (allowed.Contains("FlowDisparity"))
        yield return new KeyValuePair<string, IItem>("FlowDisparity", new Data.DoubleValue(DoubleMatrixCharacteristicCalculator.Disparity(qap.Weights)));
      if (allowed.Contains("DistanceDisparity"))
        yield return new KeyValuePair<string, IItem>("DistanceDisparity", new Data.DoubleValue(DoubleMatrixCharacteristicCalculator.Disparity(qap.Distances)));
    }
  }
}
